/*
 * Copyright (c) 2008, Noel Lemouel
 * All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. Neither the name of this project nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *    
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND THE CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * HOLDERS OR THE CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,
 * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
 * BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
 * OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
 * ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
 * TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
 * USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <sysutil.h>
#include <s29gl.h>
#include <types.h>
#include <errno.h>

/* ultra lite flash driver for s29glxxN flash series */
/* word mode only */

/* tested working with S29GL128N / S29GL256N / AM29LV640MH */

static s29gl_flash_profile_t flash_profile;

static s29gl_flash_cmd_t cmd;

int s29gl_init(int base_address)
{
    u16 cfi_ascii_QRY[3];
    u16 erase_block_region1_info[4];
    u32 i;

    flash_profile.base_address = base_address;
    
    cmd.addr0 = (volatile u16 *) (flash_profile.base_address + (0x555<<1));
    cmd.addr1 = (volatile u16 *) (flash_profile.base_address + (0xAAA<<1));
    cmd.device_ID_x01 = (volatile u16 *) (flash_profile.base_address + (0x001<<1));
    cmd.device_ID_x0e = (volatile u16 *) (flash_profile.base_address + (0x001<<1));    
    cmd.device_ID_x0f = (volatile u16 *) (flash_profile.base_address + (0x001<<1));
    cmd.manufacturer_ID = (volatile u16 *) (flash_profile.base_address + (0x000<<1));

    /* device reset */
    *cmd.addr0 = (u16)0xf0;
    
    /* CFI query mode */
    cmd.cfi_addr = (volatile u16 *) (flash_profile.base_address + (0x55<<1));
    *cmd.cfi_addr = (u16)0x98;
    
    /* read the 'QRY' string */

    /* cfi_addr = flash_profile.base_address + (0x10<<flash_profile.nbyte_mode); */
    cmd.cfi_addr = (volatile u16 *)  (flash_profile.base_address + 0x20);
    for(i=0; i<3; i++){
        cfi_ascii_QRY[i] = *cmd.cfi_addr;
        cmd.cfi_addr++;
    }
    
    if(!((cfi_ascii_QRY[0] == (u16)0x0051) &&
         (cfi_ascii_QRY[1] == (u16)0x0052) &&
         (cfi_ascii_QRY[2] == (u16)0x0059))) {
    #ifdef S29GL_INFO
      printf("%s: fatal: no CFI flash found at 0x%08x\n", __FILE__, flash_profile.base_address);
    #endif    
    return ERR_FATAL; }
    
    /* now get the mem size */
    
    /* cfi_addr = flash_profile.base_address + (0x27<<flash_profile.nbyte_mode); */
    cmd.cfi_addr = (volatile u16 *) (flash_profile.base_address + 0x4e);
    switch (*cmd.cfi_addr)
    {
        case(0x17):
            flash_profile.device_size = 64;
            break;
        case(0x18):
            flash_profile.device_size = 128;
            break;
        case(0x19):
            flash_profile.device_size = 256;
            break;
        case(0x1A):
            flash_profile.device_size = 512;
            break;
        default:
            #ifdef S29GL_INFO
               printf("%s: fatal: error reading CFI table\n", __FILE__);
            #endif
            return ERR_FATAL;
    }
    #ifdef S29GL_INFO
      printf("%s: found %dMB CFI device at 0x%08x\n",  __FILE__, (flash_profile.device_size)>>3, flash_profile.base_address);
    #endif

    /* device geometry definition */

    /* cfi_addr = flash_profile.base_address + (0x2c<<flash_profile.nbyte_mode); */
    cmd.cfi_addr = (volatile u16 *) (flash_profile.base_address + 0x58);
    if((*cmd.cfi_addr) != (u16)0x01){
      #ifdef S29GL_INFO
         printf("%s: fatal: non-uniform device not supported\n", __FILE__);
      #endif   
      return ERR_FATAL;
    }

    /* cfi_addr = flash_profile.base_address + (0x2d<<flash_profile.nbyte_mode); */
    cmd.cfi_addr = (volatile u16 *) (flash_profile.base_address + 0x5a);     
    for(i=0; i<4; i++){
      erase_block_region1_info[i] = *cmd.cfi_addr;
      cmd.cfi_addr++;
    }

    flash_profile.sector_count = (erase_block_region1_info[0] | (erase_block_region1_info[1] << 8)) + 1;
    flash_profile.sector_size = ((flash_profile.device_size) << 7) / flash_profile.sector_count;
    #ifdef S29GL_INFO
      printf("%s: geometry: %d sectors of %dkB\n", __FILE__, flash_profile.sector_count, flash_profile.sector_size);
    #endif
    /* quit CFI query mode (reset) */
    *cmd.addr0 = (u16)0xf0;

    /* read manufacturer ID */

    *cmd.addr0 = (u16)0xaa;
    *cmd.addr1 = (u16)0x55;
    *cmd.addr0 = (u16)0x90;
    
    #ifdef S29GL_INFO
      printf("%s: manufacturer ID: 0x%04x\n", __FILE__, *cmd.manufacturer_ID);
    #endif

    /* reset */
    *cmd.addr0 = (u16)0xf0;

    /* read device ID */

    *cmd.addr0 = (u16)0xaa;
    *cmd.addr1 = (u16)0x55;
    *cmd.addr0 = (u16)0x90;
    
    #ifdef S29GL_INFO
      printf("%s: device ID: 0x%04x [0x%04x|0x%04x]\n", __FILE__, *cmd.device_ID_x01, *cmd.device_ID_x0e, *cmd.device_ID_x0f);
    #endif
 
    /* reset */
    *cmd.addr0 = (u16)0xf0;

    return OK;
}

int s29gl_erase_all(void)
{
    u16 read;
    bool_t finished;

    /* should check if the flash is already empty */

    *cmd.addr0 = (u16)0xf0; /* reset */

    *cmd.addr0 = (u16)0xAA; /* erase setup */
    *cmd.addr1 = (u16)0x55;
    *cmd.addr0 = (u16)0x80;
    *cmd.addr0 = (u16)0xAA; /* chip erase */
    *cmd.addr1 = (u16)0x55; 
    *cmd.addr0 = (u16)0x10;

    *cmd.addr0 = flash_profile.base_address;

    /* disable interrupts... */
    
    finished = FALSE;
    while (finished == FALSE)
    {
        read = *cmd.addr0;
        if ((read & 0x0080) != 0)
        {
            finished = TRUE;
        }
        else if ((read & 0x0020) != 0)
        {
            read = *cmd.addr0;
            if ((read & 0x0080) != 0)
            {
                finished = TRUE;
            }
            else
            {
                return -ERR_FATAL;
            }
        }
    }

    /* enable interrupts... */
    
    *cmd.addr0 = (u16)0xf0; /* reset */

    return OK;
}

int s29gl_sector_erase(u32 sector_num)
{
    bool_t finished;

    u16 read;

    if(sector_num > flash_profile.sector_count)
        return -ERR_BAD_PARAMETER;

    cmd.sector_base = (volatile u16*) \
                      (flash_profile.base_address + (sector_num * (flash_profile.sector_size<<10)));

    if(s29gl_is_sector_empty(sector_num)) {
    #ifdef S29GL_VERBOSE
       printf("%s: sector_erase: erase not required for sector %d [base 0x%08x]\n", \
             __FILE__, sector_num, cmd.sector_base);
    #endif
       return OK;
    }
    
    *cmd.addr0 = (u16)0xf0;   /* reset */

    *cmd.addr0 = (u16)0xaa;
    *cmd.addr1 = (u16)0x55;
    *cmd.addr0 = (u16)0x80;
    *cmd.addr0 = (u16)0xaa;    
    *cmd.addr1 = (u16)0x55;
    
    *cmd.sector_base = (u16)0x30;

    /* disable interrupts... */
    
    finished = FALSE;
    while (finished == FALSE)
    {
        read = *cmd.sector_base;
        if ((read & 0x0080) != 0)
        {
            finished = TRUE;
        }
        else if ((read & 0x0020) != 0)
        {
            read = *cmd.sector_base;
            if ((read & 0x0080) != 0)
            {
                finished = TRUE;
            }
            else
            {
                #ifdef S29GL_VERBOSE
                  printf("%s: sector_erase: fatal: sector %d [base: 0x%08x] not erased\n", \
                        __FILE__, sector_num, cmd.sector_base);
                #endif
                return -ERR_FATAL;
            }
        }
     }

    /* enable interrupts... */
    
    *cmd.addr0 = (u16)0xf0;   /* reset */

    #ifdef S29GL_VERBOSE
      printf("%s: sector_erase: sector %d [base: 0x%08x] erased\n", __FILE__, sector_num, cmd.sector_base);
    #endif

    return OK;
}

bool_t s29gl_is_sector_empty(u32 sector)
{
    u32 sector_offset;

    if(sector > flash_profile.sector_count)
        return -ERR_BAD_PARAMETER;

    cmd.sector_base = (volatile u16*) \
                      (flash_profile.base_address + (sector * (flash_profile.sector_size<<10)));
    
    for(sector_offset=0; sector_offset < (flash_profile.sector_size << 9); sector_offset++) {
       if((*(cmd.sector_base + sector_offset)) != (u16)0xffff) {
            return FALSE;
       }
    }

    return TRUE;
}

int s29gl_get_sector_num(u32 address)
{

    /* flash_profile.device_size is in Mbits */

    if (address < flash_profile.base_address) return -ERR_BAD_PARAMETER;
    if (address > (flash_profile.base_address + \
                  ((flash_profile.device_size)<<17)-1)) return -ERR_BAD_PARAMETER;

    return address / (flash_profile.sector_size<<10);
}

/* raw write to flash */
/* only word mode support for the moment */

int s29gl_raw_write(u16* buf, u32 length, u32 base)
{
    bool_t finished;
    u16 read;
    u32 i;

    /* if (!flash_profile.nbyte_mode) return -ERR_FEATURE_NOT_SUPPORTED; */
    
    /*******************/

    if (base < flash_profile.base_address) return -ERR_BAD_PARAMETER;
    if ((base + length) > (flash_profile.base_address + \
                  ((flash_profile.device_size)<<17)-1)) return -ERR_BAD_PARAMETER;

    /*******************/

    cmd.word_address = (volatile u16 *) \
                       (flash_profile.base_address + base);
    
    for(i=0; i<length; i++)
    {
       if (*cmd.word_address != 0xffff) s29gl_sector_erase(s29gl_get_sector_num((u32)cmd.word_address));
       
       /* *cmd_addr0 = (u16)0xf0; */   /* reset */
       
       *cmd.addr0 = (u16)0xaa;
       *cmd.addr1 = (u16)0x55;
       *cmd.addr0 = (u16)0xa0;
       *cmd.word_address = buf[i];

       /* disable interrupts... */
       
       finished = FALSE;
       while (finished == FALSE)
       {
            read = *cmd.word_address;
            if ((read & 0x0080) == (buf[i] & 0x0080))
            {
                finished = TRUE;
            }
            else if ((read & 0x0020) == 0x0020)
            {
                read = *cmd.word_address;
                if ((read & 0x0080) == (buf[i] & 0x0080))
                {
                    finished = TRUE;
                }
                else
                {
                    #ifdef S29GL_VERBOSE
                     printf("%s: programming: fatal: write failed at 0x%08x [value: 0x%08x]", \
                           __FILE__, cmd.word_address, buf[i]);
                    #endif
                    return -ERR_FATAL;
                }
            }
       }

       cmd.word_address++;
    }

    /* enable interrupts... */
    
    return OK;
}

